freeze;

// printing
intrinsic Print( q::DiagQuadFrm )
{Print q}
    printf "(";
    d := Dimension(q);
    for i in [1..d] do 
        if i gt 1 then printf ", "; end if;
        printf "%o", q`coeffs[i];
    end for;
    printf ")";
end intrinsic;

// basic conversions
intrinsic QuadraticFormPolynomial( q::DiagQuadFrm ) -> RngPolElt
{
Converts a diagonal quadratic (a1,...,ad) form to a polynomial 
a1*x1^2 + ... + ad*xd^2;
}
    d := Dimension(q);
    P := PolynomialRing(q`base_ring, d);
    return &+ [ q`coeffs[j]*P.j^2 : j in [1..d] ];
end intrinsic;

/*
intrinsic IsCoercible( P::RngMPol, q::DiagQuadFrm ) -> BoolElt, .
{
Return whether q is coercible into a polynomial ring P and the result if so
}
    p := QuadraticFormPolynomial(q);
    return IsCoercible(P, p);
end intrinsic;
*/

intrinsic QuadraticSpace( q::DiagQuadFrm ) -> ModTupRng
{
Convert a diagonal quadratic form to a quadratic space.
This function is equivalent to QuadraticSpace(QuadraticFormPolynomial(q))
}
    return QuadraticSpace(QuadraticFormPolynomial(q));
end intrinsic;

intrinsic QuadraticFormMatrix( q::DiagQuadFrm ) -> ModAlgElt
{
Convert a diagonal quadratic form into a diagonal matrix representing
this form. 
}
    return DiagonalMatrix( q`coeffs );
end intrinsic;

// diagonalization of a quadratic space
intrinsic Diagonalization( V::ModTupFld ) -> DiagQuadFrm
{
Given a quadratic space V, this function constructs a diagonal quadratic form isometric to V.
}
    require IsPolarSpace(V): "Argument must be a quadratic space!";
    require PolarSpaceType(V) eq "quadratic space": "Argument must be a quadratic space!";
    K := BaseField(V);
    require Characteristic(K) ne 2: "Characteristic 2 is not supported";

    M := QuadraticFormMatrix(V);
    D := Diagonalization( 1/2*(M+Transpose(M)) );
    n := Ncols(D);
    return DiagonalQuadraticForm(K, [ D[j,j] : j in [1..n] ]);
end intrinsic;

intrinsic Coefficients( q::DiagQuadFrm ) -> []
{
Returns the sequence of the coefficients of q.
}
    return q`coeffs;
end intrinsic;

intrinsic ElementToSequence( q::DiagQuadFrm ) -> []
{ " } //"
    return Coefficients(q);
end intrinsic;

intrinsic Eltseq( q::DiagQuadFrm ) -> []
{ " } //"
    return Coefficients(q);
end intrinsic;


intrinsic Evaluate( q::DiagQuadFrm, v::[ RngElt ] ) -> RngElt
{
Evaluate q at v.
}
    require #v eq Dimension(q): "The dimensions do not fit!";
    C := Coefficients(q);
    K := BaseRing(q);
    require forall{ a : a in v | IsCoercible(K, a) }: "The coefficients of v are not coercible to the base field of q!";
    
    return &+ [ C[j]*v[j]^2 : j in [1..#v] ];
end intrinsic;

intrinsic Evaluate( q::DiagQuadFrm, v::ModTupFldElt ) -> RngElt
{ " } //"
    return Evaluate(q, Eltseq(v));
end intrinsic;

intrinsic Evaluate( q::DiagQuadFrm, v::ModTupRngElt ) -> RngElt
{ " } //"
    return Evaluate(q, Eltseq(v));
end intrinsic;
