freeze;

import "../real_numbers/anisotropic_part.m" : anisotropic_part_RR;
import "../complex_numbers/anisotropic_part.m" : anisotropic_part_CC;
import "../rationals/anisotropic_part.m" : anisotropic_part_QQ;
import "../finite_fields/anisotropic_part.m" : anisotropic_part_FF;
import "../number_fields/anisotropic_part.m" : anisotropic_part_NF;
import "../global_function_fields/anisotropic_part.m" : anisotropic_part_GFF;
import "../rational_global_functions/anisotropic_part.m" : anisotropic_part_RFF;

import "../real_numbers/isotropic_vector.m" : isotropic_vector_RR;
import "../complex_numbers/isotropic_vector.m" : isotropic_vector_CC;
import "../rationals/isotropic_vector.m" : isotropic_vector_QQ;
import "../finite_fields/isotropic_vector.m" : isotropic_vector_FF;
import "../number_fields/isotropic_vector.m" : isotropic_vector_NF;
import "../global_function_fields/isotropic_vector.m" : isotropic_vector_GFF;
import "../rational_global_functions/isotropic_vector.m" : isotropic_vector_RFF;


//////////////////////////////////////////////////
// AnisotropicPart

intrinsic AnisotropicPart( q::DiagQuadFrm ) -> DiagQuadFrm
{
Given a diagonal quadratic form q over a number field, this function computes its anisotropic part q_a, that is an anisotropic form s.t. q = q_a ⊥ w*<1,-1>.
}
    K := BaseRing(q);
    require IsField(K): "The base ring of K must be a field.";

    if ISA(Category(K), FldRe) then
        return anisotropic_part_RR(q);
    elif ISA(Category(K), FldCom) then
        return anisotropic_part_CC(q);
    elif ISA(Category(K), FldFin) then
        return anisotropic_part_FF(q);
    elif ISA(Category(K), FldRat) then
        return anisotropic_part_QQ(q);
    elif ISA(Category(K), FldNum) then
        return anisotropic_part_NF(q);
    elif ISA(Category(K), FldFun) then
        if IsGlobal(K) then
            return anisotropic_part_GFF(q);
        end if;
    elif ISA(Category(K), FldFunRat) then
        if IsGlobal(K) then
            return anisotropic_part_RFF(q);
        end if;
    else
        require false: "The base field is not supported!";
    end if;
    
    return false;
end intrinsic;

//////////////////////////////////////////////////
// IsotropicVector

intrinsic IsotropicVector( q::DiagQuadFrm : algorithm := "v2"  ) -> ModTupFldElt
{
Given a diagonal quadratic form q over a number field, this function computes its anisotropic part q_a, that is an anisotropic form s.t. q = q_a ⊥ w*<1,-1>.
}
    K := BaseRing(q);
    require IsField(K): "The base ring of K must be a field.";
    require IsIsotropic(q): "The given form is not isotropic.";
    
    if assigned q`isovec then
        return q`isovec;
    end if;

    if ISA(Category(K), FldRe) then
        v := isotropic_vector_RR(q);
    elif ISA(Category(K), FldCom) then
        v := isotropic_vector_CC(q);
    elif ISA(Category(K), FldFin) then
        v := isotropic_vector_FF(q);
    elif ISA(Category(K), FldRat) then
        v := isotropic_vector_QQ(q);
    elif ISA(Category(K), FldNum) then
        v := isotropic_vector_NF(q : alg := algorithm);
    elif ISA(Category(K), FldFun) then
        if IsGlobal(K) then
            v := isotropic_vector_GFF(q : alg := algorithm);
        end if;
    elif ISA(Category(K), FldFunRat) then
        if IsGlobal(K) then
            v := isotropic_vector_RFF(q : alg := algorithm);
        end if; 
    else
        require false: "The base field is not supported!";
    end if;

    q`isovec := v;   
    return v;
end intrinsic;
