freeze;

function isotropic_vector_FF_s1( q )
// isotropic vector, when -1 is a square

    K := BaseField(q);
    
    // find two coeffs in the same square class
    // it suffices to check the thirst three
    C := Coefficients(q);
    j := 1;
    if IsSquare(C[j]*C[2]) then 
        k := 2;
    elif IsSquare(C[j]*C[3]) then
        k := 3;
    else 
        j := 2;
        k := 3;
    end if;
    test, c := IsSquare(-C[j]/C[k]);
    assert test;
    
    // initialize the vector with zeros
    v := [K | 0 : c in C];
    v[j] := 1;
    v[k] := c;
    v := Vector(K, v);
    
    assert IsZero(Evaluate(q, v));
    return v;    
end function;

function isotropic_vector_FF_s2( q )
// isotropic vector when -1 is not a square

    K := BaseField(q);
    C := Coefficients(q);
    // initialize the vector with zeros
    v := [K | 0 : c in C];
    
    // do we have isotropic subform of dimension 2?
    I := [ i : i in [1..#C] | IsSquare(C[i]) ];
    J := [ j : j in [1..#C] | not IsSquare(C[j]) ];
    if not IsEmpty(I) and not IsEmpty(J) then
        i := I[1];
        j := J[1];
        test, c := IsSquare( -C[i]/C[j] );
        assert test;
        v[i] := 1;
        v[j] := c;
    elif not IsEmpty(I) then
        // the form must have dimension at least 3
        a := C[I[1]];
        b := C[I[2]];
        c := C[I[3]];
        L := ext< K | PolynomialRing(K).1^2 + c/b >;
        test, s := NormEquation(L, -a/b);
        assert test;
        s := Eltseq(s);
        v[I[1]] := 1;
        v[I[2]] := s[1];
        v[I[3]] := s[2];
    else
        // the form must have dimension at least 3
        a := C[J[1]];
        b := C[J[2]];
        c := C[J[3]];
        L := ext< K | PolynomialRing(K).1^2 + c/b >;
        test, s := NormEquation(L, -a/b);
        assert test;
        s := Eltseq(s);
        v[J[1]] := 1;
        v[J[2]] := s[1];
        v[J[3]] := s[2];
    end if;
    
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isotropic_vector_FF( q : alg := "v2" )
    assert IsIsotropic(q);
    K := BaseField(q);
    
    if Level(K) eq 1 then
        return isotropic_vector_FF_s1(q);
    else
        return isotropic_vector_FF_s2(q);
    end if;
end function;

