 freeze;

//////////////////////////////////////////////////
// IsSumOfSquares
// (just for completeness)
intrinsic IsSumOfSquares( a::FldFinElt ) -> BoolElt
{ " }
    return true;
end intrinsic;

intrinsic IsSOS( a::FldFinElt ) -> BoolElt
{ " }
    return IsSumOfSquares(a);
end intrinsic;

//////////////////////////////////////////////////
// LengthOfSOS
// (just for completeness)
intrinsic LengthOfSOS( a::FldFinElt ) -> RngIntElt
{
Given a nonzero element a, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    return IsSquare(a) select 1 else 2;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldFinElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


//////////////////////////////////////////////////
// SolveOfSOS

intrinsic SolveSumOfSquares( a::FldFinElt ) -> []
{
Let a be a sum of squares in K. Assume that the length of a is n.
This function solves the equation a = x1^2 + ... + xn^2 in K.
}

    // if a is a square, there is nothing to do
    test, s := IsSquare(a);
    if test then
        return [ s ];
    end if;
    // the base field
    K := Parent(a);
    test, s := IsSquare(K!-1);
    // if -1 is a square, a simple fomula suffices
    if test then
        c1 := (a+1)/2;
        c2 := s*(a-1)/2;
    else 
        L := ext< K | PolynomialRing(K).1^2 + 1 >;
        // generator of the multiplicative group
        //u := PrimitiveElement(L);
        // its conjugate
        //u_ := L![ U[1], -U[2] ] where U is Eltseq(u);
        //c := u*u_;
        //c := Norm(u);
        
        test, s := NormEquation(L, K!-1);
        assert test;
        s := Eltseq(s);
        test, c := IsSquare(-a);
        assert test;
        c1 := c*s[1];
        c2 := c*s[2];
    end if;
    
    assert c1^2 + c2^2 eq a;
    return [ c1, c2 ];
end intrinsic;

intrinsic SolveSOS( a::FldFinElt ) -> []
{ " } //"
    return SolveSumOfSquares(a);
end intrinsic;
