freeze;

//////////////////////////////////////////////////
// LegendreSymbol

intrinsic LegendreSymbol( a::FldFunElt, P::PlcFunElt ) -> RngIntElt
{ 
Given an element a and a place P of a function field, 
this function computes the Legendre symbol of a square-class of a over P. 
}
    K := FunctionField(P);
    require Characteristic(K) ne 2: "Implemented only for fields of odd characteristic!";
    require IsCoercible(K, a) : "Incompatible fields!";
    require IsFinite(ConstantField(K)): "The base field must be finite!";

    v := Valuation(a, P);
    if IsOdd(v) then
        return 0;
    end if;
    if v ne 0 then
        p := LocalUniformizer(P);
        a := a*p^(-v);
    end if;
    if IsSquare(Evaluate(a,P)) then
        return 1;
    end if;
    return -1;
end intrinsic;

intrinsic LegendreSymbol( a::FldFunElt, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return LegendreSymbol(a, Place(P));
end intrinsic;


//////////////////////////////////////////////////
// IsLocalSquare
intrinsic IsLocalSquare( a::FldFunElt, P::PlcFunElt ) -> BoolElt
{
Given an element a in K and a place P of a function field, 
checks whether a is a square at the completion of K at P.
}
    return LegendreSymbol(a,P) eq 1;
end intrinsic;


//////////////////////////////////////////////////
// HilbertSymbol 
// (partially based on Magma code for number fields)

intrinsic HilbertSymbol( a::FldFunElt, b::FldFunElt, P::PlcFunElt ) -> RngIntElt
{
Given a,b in a global function field and a prime p of K, 
returns the Hilbert symbol (a,b)_P.
}
    require Parent(a) eq Parent(b) : "Arguments must belong to the same field!";
    K := FunctionField(P); 
    require IsCoercible(K, a) and IsCoercible(K, b) : "Incompatible fields!";
    require not IsZero(a) and not IsZero(b): "Arguments must be non zero!";
    require IsOdd(Characteristic(K)): "Not implemented in even characteristic!";

    va := Valuation(a, P);
    vb := Valuation(b, P);
    p := LocalUniformizer(P);
    a_ := a*p^(-va);
    b_ := b*p^(-vb);
    
    // four cases two consider
    if IsEven(va) and IsEven(vb) then 
        return 1;
    elif IsEven(va) and IsOdd(vb) then
        return IsSquare(Evaluate(a_,P)) select 1 else -1;
    elif IsOdd(va) and IsEven(vb) then
        return IsSquare(Evaluate(b_,P)) select 1 else -1;
    else
        return IsSquare(Evaluate(-a_*b_,P)) select 1 else -1;
    end if;
end intrinsic;

intrinsic HilbertSymbol( a::FldFunElt, b::FldFunElt, P::RngFunOrdIdl ) -> RngIntElt
{ " } //"
    require IsPrime(P): "P must be a prime ideal!";
    return HilbertSymbol(a, b, Place(P));
end intrinsic

