freeze;

import "../helpers/varia.m" : ZZ, FF2;
import "./quadratic_forms.m" : primes_that_matter;
import "./singular_group.m" : SSingularModSquares;

forward isotropic_vector_GFF;

/////////////////////////////////////////////
//
// Isotropic vectors for various dimes

function isovec2_GFF( q )
// given a quadratic form of dimension 2, return its isotropic vector
    assert Dimension(q) eq 2;
    
    C := Coefficients(q);
    test, a := IsSquare( -C[1]*C[2] );
    assert test;
    K := BaseRing(q);
    w := Vector(K, [ C[2], a ]);
    assert IsZero(Evaluate(q, w));
    
    return w;   
end function;

function isovec3_GFF_builtin( q )
// given a quadratic form of dimension 3, return its isotropic vector
// this function uses built-in function NormEquation
    assert Dimension(q) eq 3;
    
    K := BaseRing(q);
    C := Coefficients(q);
    
    // check if there is a dimension 2 isotropic subform
    test, a := IsSquare( -C[1]*C[2] );
    if test then return Vector(K, [ C[2], a, 0 ]); end if;
    test, a := IsSquare( -C[1]*C[3] );
    if test then return Vector(K, [ C[3], 0, a ]); end if;
    test, a := IsSquare( -C[2]*C[3] );
    if test then return Vector(K, [ 0, C[3], a ]); end if;
    
    a := C[1]; b := C[2]; c := C[3];
    x := PolynomialRing(K).1;
    L := ext< K | x^2 + b/a >;
    
    test, v := NormEquation(L, -c/a);
    assert test;
    // convert to coeffs
    v := Eltseq( v[1] );
/* To jest chyba zbędne:   
    test, v3 := IsSquare( -a/c*v[1]^2 - b/c*v[2]^2 );
    assert test;

    w := Vector(K, [v[1], v[2], v3]);
*/
    w := Vector(K, v cat [1]);
    assert IsZero(Evaluate(q, w));
    
    return w;
end function;

function relative_pseudo_class_group( L )
// given a field extension L/K compute the generators of the
// pseudo-class group Cl_i(L/K). Return two sequences:
// S places in K and S_ places in L

    // retrieve K
    K := BaseField(L);
    // We will compute the relative pseudo-class group Cl_i(L/K)
    // the hard (naive) way
    DivL := DivisorGroup(L);

    // class group Cl(L)
    // mClL : Cl(L) -> Div(L)
    // ClLm : Div(L) -> Cl(L)
    ClL, mClL, ClLm := ClassGroup(L);
    ClK, mClK, ClKm := ClassGroup(K);
    // compute the image of ClK in ClL
    iClK := sub< ClL | [ ClLm( DivL ! mClK(g) ) : g in Generators(ClK) ] >;
    // now we have the ralative pseudo-class group
    Cli, kappa := ClL/iClK;
    
    // find a set of places that generate Cli
    S_ := [ mClL(g @@ kappa) : g in Generators(Cli) ];
    S_ := &cat [ Support(g) : g in S_ ];
    assert2 sub< Cli | [ kappa( ClLm(p) ) : p in S_ ]> eq Cli; 
    // find the places in K dominated by S0_
    S := [ Support(Norm(p))[1] : p in S_ ];
    
    return S, S_;
end function;

function isovec4_GFF_v1( q )
    assert Dimension(q) eq 4;
    
    K := BaseRing(q);
    C := Coefficients(q);
    
    // check if there is a dimension 3 isotropic subform
    q1 := DiagonalQuadraticForm(K, Coefficients(q)[ [2,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [0, v[1], v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, Coefficients(q)[ [1,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [v[1], 0, v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, Coefficients(q)[ [1,2,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [v[1], v[2], 0, v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, Coefficients(q)[ [1,2,3] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [v[1], v[2], v[3], 0]);
    end if;

    // two quadratic extensions
    x := PolynomialRing(K).1;
    L := ext< K | x^2 + C[2]/C[1] >;
    M := ext< K | x^2 + C[4]/C[3] >;
    DivL := DivisorGroup(L);
    DivM := DivisorGroup(M);
    
    // find the big enough set of primes
    S0 := primes_that_matter(q);
    S1, S1_ := relative_pseudo_class_group(L);
    S2, S2_ := relative_pseudo_class_group(M);
    
    // S0-class group
    // mClS0 : ClS0 -> Div(K)
    // ClS0m : Div(K) -> ClS0
    ClS0, mClS0, ClS0m := SClassGroup(Seqset(S0));
    CClS0, mCClS0 := ClS0/(2*ClS0);
    // make sure the class number is odd
    S3 := &cat [ Support(ClS0m(p @@ mCClS0)) : p in Generators(CClS0) ];
    
    S := S0 cat S1 cat S2 cat S3;
    S := Setseq(Seqset(S));
    
    // main loop
    found := false;
    // initialization of places' search
    places := PlaceEnumInit(K);
    while not found do
        // places of L dominating S
        //SL_ := &join [ Seqset(Support(DivL ! p)[1]) : p in S ];
        SL_ := Setseq(Seqset(&cat [ Support(DivL ! p) : p in S ]));
        // places of M dominating S
        //SM_ := &join [ Seqset(Support(DivM ! p)[1]) : p in S ];
        SM_ := Setseq(Seqset(&cat [ Support(DivM ! p) : p in S ]));
    
        // construct the bases of corresponding S-unit groups
        // mUS : US -> K
//         US, mUS := SUnitGroup(Seqset(S));
//         UUS, mUUS := US/(2*US);
//         B_K := [ mUS(g @@ mUUS) : g in Generators(UUS) ];
//         USL, mUSL := SUnitGroup(SL);
//         UUSL, mUUSL := USL/(2*USL);
//         B_L := [ mUSL(g @@ mUUSL) : g in Generators(UUSL) ];
//         USM, mUSM := SUnitGroup(SM);
//         UUSM, mUUSM := USM/(2*USM);
//         B_M := [ mUSM(g @@ mUUSM) : g in Generators(UUSM) ];
    
        // construct the bases of corresponding S*-singular groups
        // mEE : EE -> K
        EE_K, mEE_K := SSingularModSquares(S);
        B_K := [ mEE_K(g) : g in Generators(EE_K) ];
        EE_L, mEE_L := SSingularModSquares(SL_);
        B_L := [ mEE_L(g) : g in Generators(EE_L) ];
        EE_M, mEE_M := SSingularModSquares(SM_);
        B_M := [ mEE_M(g) : g in Generators(EE_M) ];

    
        // coordinates of a_1 in EE_K
        //a := Eltseq( mUUS(C[1] @@ mUS) );
        a := Eltseq( C[1] @@ mEE_K );
        // coordinates of -a_3 in EE_K
        //b := Eltseq( mUUS(-C[3] @@ mUS) );
        b := Eltseq( (-C[3]) @@ mEE_K );
        // find the coords of norms of all els in B_L
        //A := [ Eltseq( mUUS(Norm(l) @@ mUS) ) : l in B_L ];
        A := [ Eltseq( Norm(L!l) @@ mEE_K ) : l in B_L ];
        // find the coords of norms of all els in B_M
        //B := [ Eltseq( mUUS(Norm(m) @@ mUS) ) : m in B_M ];
        B := [ Eltseq( Norm(M!m) @@ mEE_K ) : m in B_M ];
        
        // form the system (♣)
        M := Matrix(FF2, A cat B);
        v := Vector(FF2, a) + Vector(FF2, b);
        found, V, _ := IsConsistent(M, v);
        if found then
            V := Eltseq(ChangeRing(V,ZZ));
            v1v2 := ProductRepresentation(B_L, V[1..#B_L]);
            v3v4 := ProductRepresentation(B_M, V[#B_L+1..#V]);
        
            test, c := IsSquare( C[1]*Norm( v1v2 ) / (-C[3]*Norm( v3v4 )) ); 
            assert test;

            v := Vector(K, Eltseq(v1v2) cat Eltseq(c*v3v4));
        else
            P := PlaceEnumNext(places);
            while P in S do
                P := PlaceEnumNext(places);
            end while;
            Append(~S, P);
            vprint DQF_verbosity : "Expanding S. Added P =", P;
        end if;
    end while;

    assert IsZero(Evaluate(q, v));
    return v;
end function;

function isovec4_GFF_v2( q )
// This version of the function is sometimes faster, but the results may be not so nice
    assert Dimension(q) eq 4;
    
    K := BaseField(q);
    //OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // check if there is a dimension 3 isotropic subform
    q1 := DiagonalQuadraticForm(K, C[ [2,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [0, v[1], v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [v[1], 0, v[2], v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [v[1], v[2], 0, v[3]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isovec3_GFF_builtin(q1) );
        return Vector(K, [v[1], v[2], v[3], 0]);
    end if;

    // primes to be considered
    S := primes_that_matter(q);
    
    // initialization of places' search
    places := PlaceEnumInit(K);
    // main loop
    found := false;

    while not found do
        // construct the basis of corresponding S-unit group
        // mUS : US -> K
//         US, mUS := SUnitGroup(Seqset(S));
//         UUS, mUUS := US/(2*US);
//         B_K := [ mUS(g @@ mUUS) : g in Generators(UUS) ];
        // construct the basis of corresponding S*-singular group
        // mEE : EE -> K
        EE_K, mEE_K := SSingularModSquares(S);
        B_K := [ mEE_K(g) : g in Generators(EE_K) ];
        vprint DQF_verbosity : "basis:", B_K;
   
        // two vectors to be constructed
        // 1. Hilbert symbols (a_1, a_2)
        v := [ FF2 | HilbertSymbol(C[1], C[2],p) eq -1 select 1 else 0 : p in S ];
        // 2. Hilbert symbols (-a_3, -a_4)
        w := [ FF2 | HilbertSymbol(-C[3], -C[4], p) eq -1 select 1 else 0 : p in S ];
        
        // two matrices to be constructed
        // 1. Hilbert symbols (-a_1a_2, kappa)
        V := Matrix(FF2, 
            [[ HilbertSymbol(-C[1]*C[2],kappa,p) eq -1 select 1 else 0 : kappa in B_K ] : p in S ]);
        // 2. Hilbert symbols (-a_3a_4, kappa)
        W := Matrix(FF2, 
            [[ HilbertSymbol(-C[3]*C[4],kappa,p) eq -1 select 1 else 0 : kappa in B_K ] : p in S ]);
        
        // combine the matrices 
        VW := VerticalJoin(V,W);
        // combine the three vectors
        vw := Vector(FF2, v cat w);
        
        // check if there is a solution
        found, Xi, _ := IsConsistent( Transpose(VW), vw);
        
        if found then
            vprint DQF_verbosity : "Xi =", Xi;
            // the sought element
            b := ProductRepresentation(B_K, Xi_) where Xi_ := Eltseq(ChangeRing(Xi,ZZ));
        else
            P := PlaceEnumNext(places);
            while P in S do
                P := PlaceEnumNext(places);
            end while;
            Append(~S, P);
            vprint DQF_verbosity : "Expanding S. Added P =", P;
        end if;
    end while;
    
    q1 := DiagonalQuadraticForm(K, [ -b, C[1], C[2] ]);
    q2 := DiagonalQuadraticForm(K, [  b, C[3], C[4] ]);
    w1 := isovec3_GFF_builtin( q1 );
    w2 := isovec3_GFF_builtin( q2 );
    
    v := [ w1[2]/w1[1], w1[3]/w1[1], w2[2]/w2[1], w2[3]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);
end function;



function isovec5_GFF( q : alg := "v2"  )
    assert Dimension(q) eq 5;
    
    K := BaseField(q);
    //OK := MaximalOrder(K);
    C := Coefficients(q);
    
    // check if there is a dimension 4 isotropic subform
    q1 := DiagonalQuadraticForm(K, C[ [2,3,4,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_GFF(q1 : alg := alg) );
        return Vector(K, [0, v[1], v[2], v[3], v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,3,4,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_GFF(q1 : alg := alg) );
        return Vector(K, [v[1], 0, v[2], v[3], v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,4,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_GFF(q1 : alg := alg) );
        return Vector(K, [v[1], v[2], 0, v[3], v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3,5] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_GFF(q1 : alg := alg) );
        return Vector(K, [v[1], v[2], v[3], 0, v[4]]);
    end if;
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3,4] ]);
    if IsIsotropic(q1) then
        v := Eltseq( isotropic_vector_GFF(q1 : alg := alg) );
        return Vector(K, [v[1], v[2], v[3], v[4], 0]);
    end if;

    // two subforms
    q1 := DiagonalQuadraticForm(K, [C[1], C[2], C[3]]);
    q2 := DiagonalQuadraticForm(K, [C[4], C[5]]);
    
    // primes to be considered
    T := primes_that_matter(q);
    
    // initialization of places' search
    places := PlaceEnumInit(K);
    // main loop
    found := false;

    while not found do
        // construct the basis of corresponding T-unit group
        // mUS : US -> K
        US, mUS := SUnitGroup(Seqset(T));
        UUS, mUUS := US/(2*US);
        B_K := [ mUS(g @@ mUUS) : g in Generators(UUS) ];
        vprint DQF_verbosity : "basis:", B_K;
    
        // subset of T of the primes that *really matter*
        S := [ p : p in T | not IsIsotropic(q1, p) or not IsIsotropic(q2, p) ];
    
        // generators of local square-class groups
        // and elements c s.t. q1 represents c and q2 represents -c (at P)
        HH := [];
        cc := [];
        for P in S do
            V, mV := LocalMultiplicativeGroupModSquares( P );
            Append(~HH, [ mV(v) : v in Generators(V) ]);
            
            for v in V do
                ci := mV(v);
                if IsIsotropic( DiagonalQuadraticForm(K, [-ci, C[1], C[2], C[3]]) )
                    and IsIsotropic( DiagonalQuadraticForm(K, [ci, C[4], C[5]]) ) then
                    Append(~cc, ci);
                    break;
                end if;
            end for;
        end for;
        assert #HH eq #cc;
        
        // vector beta
        beta := [ FF2 | (b eq -1) select 1 else 0 : b in beta_ ] where 
            beta_ is &cat [ [ HilbertSymbol(cc[i], h, S[i]) : h in HH[i] ] : i in [1..#S] ];
        
        // matrix B
        // Hilbert symbols
        B_ := &cat [ 
            // this is a single block B_i
            [
                [ 
                    HilbertSymbol(h, kappa, S[i]) eq -1 select 1 else 0 
                    : kappa in B_K
                ]
                : h in HH[i]
            ]
            : i in [1..#S] 
        ];
        B := Matrix(FF2, B_);
        
        // check if there is a solution
        found, Xi, _ := IsConsistent( Transpose(B), beta);
        
        if found then
            vprint DQF_verbosity : "Xi =", Xi;
            // the sought element
            c := ProductRepresentation(B_K, Xi_) where Xi_ := Eltseq(ChangeRing(Xi,ZZ));
        else
            P := PlaceEnumNext(places);
            while P in S do
                P := PlaceEnumNext(places);
            end while;
            Append(~S, P);
            vprint DQF_verbosity : "Expanding S. Added P =", P;
        end if;       
    end while;
    
    // find isotropic vector of <-c> + q1
    w1 := Eltseq( isotropic_vector_GFF(DiagonalQuadraticForm(K, [-c]) + q1 : alg := alg) );
    // find isotropic vector of <c> + q2
    w2 := Eltseq( isovec3_GFF_builtin(DiagonalQuadraticForm(K, [c]) + q2) );

    v := [ w1[2]/w1[1], w1[3]/w1[1], w1[4]/w1[1], w2[2]/w2[1], w2[3]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isovec6_GFF( q : alg := "v2" )
    assert Dimension(q) eq 6;
    
    K := BaseField(q);
    C := Coefficients(q);
    
    // split q at half
    q1 := DiagonalQuadraticForm(K, C[ [1,2,3] ]);
    q2 := DiagonalQuadraticForm(K, [ -C[4],-C[5],-C[6] ]);
    
    if IsIsotropic(q1) then
        w := Eltseq( isotropic_vector_GFF(q) );
        v := Vector(K, w cat [0,0,0]);
        return v;
    end if;
    if IsIsotropic(q2) then
        w := Eltseq( isotropic_vector_GFF(q) );
        v := Vector(K, [0,0,0] cat w);
        return v;
    end if;
    
    T := primes_that_matter(q);
    S := [p : p in T | (not IsIsotropic(q1, p)) or (not IsIsotropic(q2, p)) ];
    
    cc := [];
    d1 := Determinant(q1);
    d2 := Determinant(q2);
    for p in S do
        V, mV := LocalMultiplicativeGroupModSquares( p );
        for v in V do
            ci := v @@ mV;
            if not IsLocalSquare(-ci*d1,p) and not IsLocalSquare(-ci*d2,p) then
                Append(~cc, ci);
                break;
            end if;
        end for;
    end for;
    assert #cc eq #S;
    
    c := ChineseRemainderTheorem( S, cc, [1 : p in S] );
    // find isotropic vector of <-c> + q1
    q1_ := DiagonalQuadraticForm(K, [-c]) + q1;
    w1 := Eltseq( isotropic_vector_GFF(q1_ : alg := alg) );
    // find isotropic vector of <c> + q2
    q2_ := DiagonalQuadraticForm(K, [-c]) + q2;
    w2 := Eltseq( isotropic_vector_GFF(q2_ : alg := alg) );

    v := [ w1[2]/w1[1], w1[3]/w1[1], w1[4]/w1[1], w2[2]/w2[1], w2[3]/w2[1], w2[4]/w2[1] ];
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isovec7up_GFF( q : alg := "v2" )
// Computes an isotropic vector of a quadratc form of dimension 7 or above over a
// non-real number field
    assert Dimension(q) gt 6;
    
    K := BaseField(q);
    C := Coefficients(q);
    q_ := DiagonalQuadraticForm(K, C[ [1..6] ]);
    w := Eltseq( isotropic_vector_GFF(q_ : alg := alg) );
    v := w cat [ 0 : j in [7..Dimension(q)] ];
    
    assert IsZero(Evaluate(q, v));
    return Vector(K, v);    
end function;

function isotropic_vector_GFF( q : alg := "v2" )
    assert IsIsotropic(q);
    K := BaseField(q);
    
    d := Dimension(q);

    C := Coefficients(q);
    for j in [1..d] do
        q_ := DiagonalQuadraticForm(K, C[ [1..j-1] cat [j+1..d] ]);
        if IsIsotropic(q_) then
            w := Eltseq(isotropic_vector_GFF(q_ : alg := alg));
            v := Vector(K, w[ [1..j-1] ] cat [0] cat w[ [j..d-1] ]);
            assert IsZero(Evaluate(q, v));
            return Vector(K, v);    
        end if;
    end for;
    
    case d :
    when 2 : return isovec2_GFF(q);
    when 3 : return isovec3_GFF_builtin(q);
    when 4 : return alg eq "v1" select isovec4_GFF_v1(q) else isovec4_GFF_v2(q);
    when 5 : return isovec5_GFF(q : alg := alg);
    when 6 : return isovec6_GFF(q : alg := alg);
    else   : return isovec7up_GFF(q : alg := alg);
    end case;
end function;

