freeze

import "../helpers/varia.m" : ZZ, FF2;



function is_S_singular( a, S )
// Checks if a is S-singular
    P, o := Support(Divisor(a));
    P := [ P[j] : j in [1..#P] | IsOdd(o[j]) ];
    return forall{ p : p in P | p in S };
end function;

function point_basis( K  : S := {@ @} ) 
    // Given a function field K, this function returns
    // a set P of primes such that { [p] + 2PicX : p in P } is a basis
    // of PicX/2PicX 
    // P will be disjoint with S, if S is provided
    
    // Picard group, psi: Div->Pic, phi:Pic->Div
    PicX, phi, psi := ClassGroup(K);
    PPicX, kappa := PicX/(2*PicX);
    n := #Generators(PPicX);
    // successively build the basis
    P := {@ @};
    G := sub< PPicX | >;
    // brute force search
    places := PlaceEnumInit(K);
    while G ne PPicX do
        p := PlaceEnumNext(places);
        if p in S then
            continue;
        end if;
        p_ := kappa(psi(1*p));
        if not p_ in G then
            Include(~P, p);
            G := sub< PPicX | G, p_ >;
        end if;       
    end while;

    return P;
end function;

function coords_in_EES(a, P, S)
// for a basis of EE_S of a form B\cup L
// where B is a basis of EE compatible with P
// and L is a basis associated to S
// this function returns coordinates of element a

    // coordinates with respect to L/S-part of basis
    e := [ Valuation(a,p) mod 2 : p in S ];
    // coordinates w.r.t B/P part of basis
    s := [ FF2 | IsLocalSquare(a, p) select 0 else 1 : p in P ];
    return s cat e;
end function;

function maximal_linearly_independent(K, S)
// Given a sequence of places, this funtion returns
// a sequence of indices s.t. S[J] is a maximal subset of S
// linearly independent in PicX/2PicX
// B is a basis of EE compatible to some basis of PicX/2PicX

    // Picard group, psi: Div->Pic, phi:Pic->Div
    PicX, phi, psi := ClassGroup(K);
    PPicX, kappa := PicX/(2*PicX);
    n := #Generators(PPicX);
    
    J := [];
    G := sub< PPicX | >;
    for j in [1..#S] do
        s := kappa(psi(1*S[j]));
        if not s in G then
            Append(~J, j);
            G := sub< PPicX | G, s >;
        end if;       
    end for;
    return J;
end function;


function SSingularModSquares( S )
// given a sequence S of places of K, this function
// construct the group of S-singular square classes together
// with a map form this group to K
// This is an implementation of algorithm from Magma manual

    K := FunctionField(Universe(S));

    ClS, mClS, ClSm := SClassGroup(Seqset(S));
    // find places generating ClS/(2*ClS)
    S0 := &join [ Seqset(Support(mClS(g @@ kappa))) : g in Generators(ClS2) ]
        where ClS2, kappa is ClS/(2*ClS);
    // group of units of enlarged S
    US, mUS := SUnitGroup(Seqset(S) join S0);
    US2, kappa := US/(2*US);
    // find the subgroup equal EE_S
    if IsEmpty(S0) then
        EES := US2;
    else
        G := AbelianGroup( [2 : p in S] );
        h := hom< 
            US2->G | 
            [ [ Valuation(mUS(g @@ kappa), p) : p in S0 diff Seqset(S) ] : g in Generators(US2) ] 
        >;
        // group of S-singular square classes
        EES := Kernel(h);
    end if;
    
    // find a basis of PicX/2PicX disjoint with S
    P := point_basis(K : S := S);
    // find indices a maximal linearly independent subset of S
    I := maximal_linearly_independent(K, S);
    SI := S[I];                                 // linearly independent
    SD := S[[j : j in [1..#S] | not j in I]];   // linearly dependent
    
    // basis of EES
    B := [ mUS(x @@ kappa) : x in Generators(EES) ];
    // matrix of coordinates of B with respect to the other basis
    M := Matrix(FF2, [ coords_in_EES(b, P, SD) : b in B ]);
    N := M^-1;
    
    function toEES( f )
        error if IsZero(f), "Not a singular element!";
        error if not is_S_singular(f, S), "Not a singular element!";
        
        // find the coords of f wrt the "other" basis
        v := Vector(FF2, coords_in_EES(f, P, SD));
        // convert ot "our" basis
        w := v*N;
        return EES ! Eltseq(w);
    end function;
    // map EES -> K^*
    mEES := map< EES->K | x :-> mUS(x @@ kappa), y :-> toEES(y) >;
    
    return EES, mEES;
end function;
