freeze;

import "../helpers/varia.m" : FF2, ZZ;

intrinsic RandomElement( K::FldFun : size := 0 ) -> FldNumElt
{
Returns a random element a of K.
}
    require IsGlobal(K) : "Unsupported field!";
    F := ConstantField(K);
    degK := Degree(K);
    if IsZero( size ) then
        size := degK;
    end if;
    
    Fx := FunctionField(F);
    a := K! [ Fx! [ Random(F) : j in [1..size] ] : i in [1..degK] ];
    return a;
end intrinsic;

intrinsic RandomDiagonalQuadraticForm( K::FldFun, d::RngIntElt : size := 0 ) -> DiagQuadFrm
{ 
Construct a random quadratic form of dimension d. 
}
    require d gt 0 : "Dimension must be positive!";
    q := DiagonalQuadraticForm(K, [0]);
    while IsDegenerate( q ) do  
        q := DiagonalQuadraticForm(K, [ RandomElement(K : size := size) : j in [1..d] ]);
    end while;
    return q;
end intrinsic;

intrinsic RandomDQF( K::FldFun, d::RngIntElt : size := 0 ) -> DiagQuadFrm
{ " } //"
    return RandomDiagonalQuadraticForm(K, d : size := size);
end intrinsic;

intrinsic LocalMultiplicativeGroupModSquares( p::PlcFunElt ) -> ModFld, Map
{
Let p be a place of a global function field K. The local square class group K^*/K^*2 is isomorphic to Klein 4-group and so it is a vectors space V of dimension 2 over GF(2). This function returns this vector space together with a map from K to V.
}
    K := FunctionField(p);
    V := VectorSpace(FF2, 2);
    // non-square unit
    Kp, mKp := ResidueClassField(p);
    u := Nonsquare(Kp) @@ mKp;
    // uniformizer
    pi := UniformizingElement(p);
    
    function toV( a )
        error if IsZero(a), "Element not in the domain of the map!";
        
        if IsEven(Valuation(a, p)) then
            if IsLocalSquare(a, p) then 
                return Zero(V);
            else
                return u;
            end if;
        else
            if IsLocalSquare(a*pi, p) then
                return pi;
            else
                return u*pi;
            end if;
        end if;
    end function;
    
    function fromV(a)
        a_ := ChangeUniverse(Eltseq(a), ZZ);
        return u^(a_[1])*pi^(a_[2]);
    end function;

    return V, map< K->V | x :-> toV(x), y :-> fromV(y) >;
end intrinsic;
