 freeze;
 
declare attributes FldNum: sos1;    // decomposition of -1 into sum of squares

import "varia.m" : initialize_prime_iterator, next_prime;
import "../helpers/varia.m" : ZZ, FF2;

//////////////////////////////////////////////////
// Helper functions for SolveSOS

function ssos_len2( a )
// Given an element of length 2, this function returns c1, c2
// such that c1^2 + c2^2 = a
    K := Parent(a);
    
    test, i := IsSquare(K!-1);
    if test then
        c1 := (a+1)/2;
        c2 := i*(a-1)/2;
    else
        P := PolynomialRing(K);
        L := ext< K | P.1^2 + 1 >;
        test, v := NormEquation(L, a : Nice := true);
        assert test;
        // convert to coeffs
        v := Eltseq( v[1] );
        c1 := v[1];
        c2 := v[2];
    end if;
    
    assert c1^2 + c2^2 eq a;
    return [ c1, c2 ];
end function;

function ssos_len3_special( a )
// Given an element of length 3, this function returns c1, c2, c3
// such that c1^2 + c2^2 + c3^2 = a
// This is a version for field of level s(K) == 2.
    K := Parent(a);
    assert Level(K) eq 2;

/*    
    if assigned K`sos1 then
        d1 := K`sos1[1];
        d2 := K`sos1[2];
    else
        P := PolynomialRing(K);
        L := ext< K | P.1^2 + 1 >;
        test, v := NormEquation(L, a : Nice := true);
        assert test;
        K`sos1 := Eltseq(v[1]);
        d1 := K`sos1[1];
        d2 := K`sos1[2];   
    end if;
*/
        P := PolynomialRing(K);
        L := ext< K | P.1^2 + 1 >;
        test, v := NormEquation(L, K!-1 : Nice := true);
        assert test;
        v := Eltseq(v[1]);
        d1 := v[1];
        d2 := v[2];   

    c := [ (a+1)/2, d1*(a-1)/2, d2*(a-1)/2 ];

    assert c[1]^2 + c[2]^2 + c[3]^2 eq a;
    return c;
end function;

function ssos_len3_general( a )
// Given an element of length 3, this function returns c1, c2, c3
// such that c1^2 + c2^2 + c3^2 = a
// This is a version for field of level s(K) != 1, 2.
    K := Parent(a);
    OK := MaximalOrder(K);
    
    // dyadic primes
    S := { s[1] : s in Factorization(2*OK) };
    // nondyadic primes where a has odd order
    S join:= { s[1] : s in Factorization(a*OK) | IsOdd(s[2]) };
    // convert into a sequence
    S := Setseq(S);
    
    // real places of K
    Soo := RealPlaces(K);
    
    // initialization of places' search
    it := initialize_prime_iterator(K);
    // main loop
    found := false;
    while not found do
        // basis of the group of S-singular elements
        // E_S group of singular elements 
        // psi : OK -> E_S
        E_S, psi := pSelmerGroup(2, Set(S) : Nice);
        B_ := [ K! (g @@ psi) : g in Generators(E_S) ];
        vprint DQF_verbosity : "basis:", B_;
    
        // three vectors to be constructed
        u := [ FF2 | 1 : i in Soo ];
        v := [ FF2 | HilbertSymbol(K!-1,K!-1,p) eq -1 select 1 else 0 : p in S ];
        w := [ FF2 | 0 : i in S ];
        
        // three matrices to be constructed
        A := Matrix(FF2, [[ a eq -1 select 1 else 0 : a in RealSigns(kappa) ] : kappa in B_ ]);
        A := Transpose(A);
        B := Matrix(FF2, 
            [[ HilbertSymbol(K!-1,kappa,p) eq -1 select 1 else 0 : kappa in B_ ] : p in S ]);
        C := Matrix(FF2, 
            [[ HilbertSymbol(a,kappa,p) eq -1 select 1 else 0 : kappa in B_ ] : p in S ]);
        
        // combine the matrices 
        // no idea why it doesn't work
        //ABC := VerticalJoin([A,B,C]);
        // using a dirty hack instead
        ABC := VerticalJoin(VerticalJoin(A,B),C);
        // combine the three vectors
        uvw := Vector(FF2, u cat v cat w);
        
        // check if there is a solution
        found, Xi, _ := IsConsistent( Transpose(ABC), uvw);
        vprint DQF_verbosity : "Xi =", Xi;
        
        if found then
            // the sought element
            b := ProductRepresentation(B_, Xi_) where Xi_ := Eltseq(ChangeRing(Xi,ZZ));
        else
            it, P := next_prime(it);
            while P in S do
                it, P := next_prime(it);
            end while;
            Append(~S, P);
        end if;
    end while;
    
    P<x> := PolynomialRing(K);
    L := ext< K | x^2 + 1 >;
    test, v := NormEquation(L, -b : Nice := true);
    assert test;
    v := Eltseq(v[1]);
    d1 := v[1];
    d2 := v[2];
    
    M := ext< K | x^2 - a >;
    test, w := NormEquation(M, b : Nice := true);
    assert test;
    w := Eltseq(w[1]);
    d3 := w[1];
    d4 := w[2];
    
    assert not IsZero(d4);
    c := [ d1/d4, d2/d4, d3/d4 ];
    assert c[1]^2 + c[2]^2 + c[3]^2 eq a;
    return c;
end function;

function prescribedHS( a, p, v )
// returned a random element b st. (a,b)_p == v

// NOT SURE IF THIS WILL ALWAYS WORK PROPERLY !!!

    assert (v eq 1) or (not IsLocalSquare(a, p));
    
    K := Parent(a);
    degK := Degree(K);
    b := K!1;
    n := Ceiling(Abs(AbsoluteNorm(a)));
    c := n;
    while IsLocalSquare(b, p) or HilbertSymbol(a,b,p) ne v do
        n -:= 1;
        if IsZero(n) then
            c *:= 2;
            n := c;
        end if;
        b := K![ Random(-c, c) : j in [1..degK] ];
    end while;
    return b;
end function;

function len4_helper( a, d )
// a is an element of a number field K
// d is a dyadic prime
// This function returns h s.t. there is g and the the Hilbert symbols
// (a, g)_d = 1 && (a, h)_d = -1

    return prescribedHS( prescribedHS(a,d,1), d, -1);
end function;

function ssos_len4( a )
// Given an element of length 4, this function returns c1, c2, c3, c4
// such that c1^2 + c2^2 + c3^2 + c4^2 = a
    K := Parent(a);
    OK := MaximalOrder(K);
    
    // dyadic primes
    S := { s[1] : s in Factorization(2*OK) };
    // dyadic primes of odd degree over QQ2
    D := [ d : d in S | IsOdd(InertiaDegree(d)*RamificationIndex(d)) ];
    // nondyadic primes where a has odd order
    S join:= { s[1] : s in Factorization(a*OK) | IsOdd(s[2]) };
    // convert into a sequence
    S := Setseq(S);
    
    // Find h_i's 
    H := [ len4_helper(a, d) : d in D ]; 
    
    // real places of K
    Soo := RealPlaces(K);
    
    // initialization of places' search
    it := initialize_prime_iterator(K);
    // main loop
    found := false;
    while not found do
        // basis of the group of S-singular elements
        // E_S group of singular elements 
        // psi : OK -> E_S
        E_S, psi := pSelmerGroup(2, Set(S) : Nice);
        B_ := [ K! (g @@ psi) : g in Generators(E_S) ];
        vprint DQF_verbosity : "basis:", B_;
    
        // three vectors to be constructed
        u := [ FF2 | 1 : r in Soo ];
        v := [ FF2 | 1 : d in D ];
        w := [ FF2 | 0 : s in S ];
        
        // three matrices to be constructed
        A := Matrix(FF2, [[ a eq -1 select 1 else 0 : a in RealSigns(kappa) ] : kappa in B_ ]);
        A := Transpose(A);
        B := Matrix(FF2, [
            [ HilbertSymbol(H[j],kappa,D[j]) eq -1 select 1 else 0 : kappa in B_ ] 
            : j in [1..#D] 
        ]);
        C := Matrix(FF2, 
            [[ HilbertSymbol(a,kappa,p) eq -1 select 1 else 0 : kappa in B_ ] : p in S ]);
        
        // combine the matrices 
        // no idea why it doesn't work
        //ABC := VerticalJoin([A,B,C]);
        // using a dirty hack instead
        ABC := VerticalJoin(VerticalJoin(A,B),C);
        // combine the three vectors
        uvw := Vector(FF2, u cat v cat w);
        
        // check if there is a solution
        found, Xi, _ := IsConsistent( Transpose(ABC), uvw);
        vprint DQF_verbosity : "Xi =", Xi;
        
        if found then
            // the sought element
            b := ProductRepresentation(B_, Xi_) where Xi_ := Eltseq(ChangeRing(Xi,ZZ));
        else
            it, P := next_prime(it);
            while P in S do
                it, P := next_prime(it);
            end while;
            Append(~S, P);
        end if;
    end while;
    
    // s(K) cannot be in {1,2}, since there exists a 4-square element
    // so we know it is the general function to be called
    d1_3 := ssos_len3_general( -b );
    
    P<x> := PolynomialRing(K);
    M := ext< K | x^2 - a >;
    test, w := NormEquation(M, b : Nice := true);
    assert test;
    w := Eltseq(w[1]);
    
    assert not IsZero(w[2]);
    c := [ di/w[2] : di in d1_3 ] cat [ w[1]/w[2] ];
    assert c[1]^2 + c[2]^2 + c[3]^2 + c[4]^2 eq a;
    return c;
end function;


intrinsic SolveSumOfSquares( a::FldNumElt ) -> []
{
Let a be a sum of squares in K. Assume that the length of a is n.
This function solves the equation a = x1^2 + ... + xn^2 in K.
}
    require IsSumOfSquares(a) : "The argument must be a sum of squares!";
    
    test, sqrt := IsSquare(a);
    if test then
        return [ sqrt ];
    end if;
    
    l := LengthOfSOS(a);
    if l eq 2 then 
        return ssos_len2(a);
    elif l eq 3 then
        K := Parent(a);
        sK := Level(K);
        if sK eq 2 then
            return ssos_len3_special(a);
        else
            return ssos_len3_general(a);
        end if;
    else 
        return ssos_len4(a);
    end if;
end intrinsic;

intrinsic SolveSOS( a::FldNumElt ) -> []
{ " } //"
    return SolveSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// IsSumOfSquares

intrinsic IsSumOfSquares( a::FldNumElt ) -> BoolElt
{
Returns true iff a is a sum of squares
}
    //return IsTotallyPositive(a);
    return forall{ s : s in RealSigns(a) | s eq 1 };
end intrinsic;

intrinsic IsSOS( a::FldNumElt ) -> BoolElt
{ " } //"
    return IsSumOfSquares(a);
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

function length_nondyadic( a, P )
// Given a nonzero element a of a number field K and a nondyadic place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  
    
    // the residue field
    Kp := ResidueClassField(P);
    // and valuation
    v := Valuation(a,P);
    // two cases to consider
    if IsOdd(v) then
        // possible values are 2 and 3
        q := DiagonalQuadraticForm(Kp, [Kp!1, Kp!1]);
        return IsIsotropic(q) select 2 else 3;
    else
        // possible values are 1 and 2
        return IsLocalSquare(a, P) select 1 else 2;
    end if;
end function;

function length_dyadic( a, P )
// Given a nonzero element a of a number field K and a dyadic place P, 
// this fuction determines the minimal number of summands needed to express 
// a as a sum of squares in a completion K_P.  

    K := Parent(a);
    // if it is a square, the answer is trivial
    if IsLocalSquare(a, P) then
        return 1;
    end if;
    // now possible values are 2, 3, 4
    // check for 2
    if HilbertSymbol(K!-1, a, Ideal(P)) eq 1 then
        return 2;
    end if;
    // the two possibilities left are 3 and 4
    // check 3
    if HilbertSymbol(K!-1, K!-1, Ideal(P)) eq 1 or not IsLocalSquare(-a, P) then
        return 3;
    end if;
    // now the only choice is 4
    return 4;
end function;

intrinsic LengthOfSOS( a::FldNumElt ) -> RngIntElt
{
Given a nonzero element a of a number field K, this function determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares
}
    require not IsZero(a): "The argument must be nonzero!";
    // start with trivial cases
    if not IsSOS(a) then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // the possible values now are 2, 3, 4
    K := Parent(a);
    OK := MaximalOrder(K);
    // find all the primes of K that matter for a
    even_primes := [ P[1] : P in Decomposition(K, 2) ];
    //odd_primes :=  [ P[1] : P in Factorization(a*OK) ];
    odd_primes, _ := [ P : P in Support(Divisor(a)) | P notin even_primes ];
    vprint DQF_verbosity : "Primes to be tested:", even_primes, odd_primes;
    
    vprint DQF_verbosity : "Lengths at dydadic primes:", [ length_dyadic(a, P) : P in even_primes ];
    m := Max([ length_dyadic(a, P) : P in even_primes ]);
    vprint DQF_verbosity : "m = ", m;
    vprint DQF_verbosity : "Lengths at non-dydadic primes:", [ length_dyadic(a, P) : P in odd_primes ];
    n := IsEmpty(odd_primes) select -1 else Max([ length_nondyadic(a, P) : P in odd_primes ]);
    vprint DQF_verbosity : "n =", n;
    l := Max([2, m, n]);
    return l;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldNumElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;
