freeze;

import "../helpers/varia.m" : QQ;

// helper structure to iterate over primes of K
PRIME_ITERATOR := recformat< p : Integers(), j : Integers(), f, K >;

function initialize_prime_iterator( K )
// This function initializes an iterator for non-dyadic primes of K
// It returns a PRIME_ITERATOR object
    return rec< PRIME_ITERATOR | p := 2, j := 0, f := [], K := K >;
end function;

function next_prime( i )
// Given a PRIME_ITERATOR object i, this function returns 
// "the next" prime of K.
    if i`j ge #i`f then
        i`p := NextPrime(i`p);
        i`f := [ Ideal(P[1]) : P in Decomposition(i`K, i`p) ];
        i`j := 0;
    end if;
    i`j +:= 1;
    return i, i`f[i`j]; 
end function;

function positive_approximation( S, K, L )
// Given a finite sequence: 
// S = [ p1, ... , pn ] of non-archimedean primes 
// K = [ k1, ... , kn ] of positive integers
// L = [ λ1, . . . , λn ] 
// this function constructs a *totally positive* element α ∈ K 
// such that α ≡ λi ( mod pi^ki ) for every i ≤ n.

// This is an implementation of [KR202?, Algorithm 6]

    assert #S eq #K and #S eq #L;
    
    
    // the base field
    K := NumberField(Universe(L));
    // number of elements
    n := #S;

    beta := ChineseRemainderTheorem( L, [ S[i]^K[i] : i in [1..n] ] );
    // rational primes dominated by places in S
    // strange syntax to make elements unique
    S_ := Setseq({@  MinimalInteger(s) : s in S @});
    // for every p in S_ find the maximal exponent ki for a primes in S
    // above p
    S_ := [ p^Max([K[i] : i in [1..n] | K!p in S[i]]) : p in S_ ];
    s := &* S_;
    
    t := 1 + Ceiling( Max([ Evaluate(-beta, sigma) : sigma in RealPlaces(K) ])/s );
    
    return beta + t*s;
end function;

// Random element of a number field
intrinsic RandomElement( K::FldNum : size := 0 ) -> FldReElt
{
Returns a random element a of K. 
}
    degK := Degree(K);
    a := K! [ RandomElement(QQ : size := size) : i in [1..degK] ];
    return a;
end intrinsic;

intrinsic RandomDiagonalQuadraticForm( K::FldNum, d::RngIntElt : size := 0 ) -> DiagQuadFrm
{ 
Construct a random quadratic form of dimension d. 
}
    require d gt 0 : "Dimension must be positive!";
    q := DiagonalQuadraticForm(K, [0]);
    while IsDegenerate( q ) do  
        q := DiagonalQuadraticForm(K, [ RandomElement(K : size := size) : j in [1..d] ]);
    end while;
    return q;
end intrinsic;

intrinsic RandomDQF( K::FldNum, d::RngIntElt : size := 0 ) -> DiagQuadFrm
{ " } //"
    return RandomDiagonalQuadraticForm(K, d : size := size);
end intrinsic;

