freeze;

//////////////////////////////////////////////////
// IsFormallyReal
// just for completeness
intrinsic IsFormallyReal( K::FldRat ) -> BoolElt
{
Returns true iff K is formally real (i.e. -1 is not a sum of squares)
}
    return true;
end intrinsic;

intrinsic IsReal( K::FldRat ) -> BoolElt
{ " } //"
    return true;
end intrinsic;


//////////////////////////////////////////////////
// Level
// just for completeness
intrinsic Level( K::FldRat ) -> RngIntElt
{    
This function comutes the level s(K) of a number field K,
i.e. a lowest number s such that -1 is a sum of s squares
it returns 0 if K is formally real.
}
    return 0;
end intrinsic;


//////////////////////////////////////////////////
// PythagorasNumber
// just for completeness
intrinsic PythagorasNumber( K::FldRat ) -> RngIntElt
{
This function computes the Pythagoras number P(K) of a number field K,
i.e. the lowest number p such that every sum of squares in K
is a sum of p squares.
}
    return 4;
end intrinsic;

intrinsic PythagorasElement( K::FldRat ) -> FldNumElt
{
Returns an element a of K whose length equals the Pythagoras 
number P(K) of K.
}
    return 7;
end intrinsic;

//////////////////////////////////////////////////
// Height
// just for completeness
intrinsic Height( K::FldRat ) -> RngIntElt
{
This function computes the height h(K) of a number field K,
i.e. exp( Wt(K) ), where Wt(K) is denotes the torsion subgroup
of the Witt ring of K..
}
    return 4;
end intrinsic;


//////////////////////////////////////////////////
// AreWittEquivalent
// just for completeness
intrinsic AreWittEquivalent( K::FldRat, L::FldRat ) -> BoolElt
{
Given two number fields K and L, this function
return true if their Witt rings are isomorphic and false otherwise.
}    
    return true;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldNum ) -> BoolElt
{
Given two number fields K and L, this function
return true if their Witt rings are isomorphic and false otherwise.
}    
    return AbsoluteDegree(L) eq 1;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldFun[FldFin] ) -> BoolElt
{ " } //"
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldFunRat ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

intrinsic AreWittEquivalent( K::FldRat, L::FldFun ) -> BoolElt
{ " } //"
    require ISA(Category(ConstantField(L)), FldFin) : "Not implemented for these types of fields!";
    // Use Szymiczek (1991), Theorem 1.3
    return false;
end intrinsic;

