freeze;

import "../helpers/varia.m" : QQ, ZZ;


//////////////////////////////////////////////////
// IsSumOfSquares
// (just for completeness)
intrinsic IsSumOfSquares( a::FldRatElt ) -> BoolElt
{
Returns true iff a is a sum of squares
}
    return a ge 0;
end intrinsic;

intrinsic IsSumOfSquares( a::RngIntElt ) -> BoolElt
{ " } //"
    return a ge 0;
end intrinsic;

intrinsic IsSOS( a::FldRatElt ) -> BoolElt
{ " } //"
    return a ge 0;
end intrinsic;

intrinsic IsSOS( a::RngIntElt ) -> BoolElt
{ " } //"
    return a ge 0;
end intrinsic;


//////////////////////////////////////////////////
// LengthOfSOS

intrinsic LengthOfSOS( a::RngIntElt ) -> RngIntElt
{
Given a nonzero integer a, this fuction determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares.
}
    require not IsZero(a): "The argument must be nonzero!";
    
    // start with trivial cases
    if a lt 0 then
        return 0;
    end if;
    if IsSquare(a) then
        return 1;
    end if;
    // special case
    if a eq 2 then
        return 2;
    end if;
    
    // check if a is a sum of 3 squares
    // write a in a form a = 4^v'*a'
    v := Valuation(a, 2);
    v_ := Floor( v/2 );
    a_ := ZZ ! ( a*4^(-v_) );
    // use Lagrange 3-square theorem
    if (a_ mod 8) eq 7 then
        return 4;
    end if;
    
    // now a_ (hence a) is a sum of 2 or 3 squares
    // compute the radical of a
    m, _ := SquarefreeFactorization(a);
    // use Lagrange 2-square theorem
    sq2 := forall{ p : p in PrimeDivisors(m) | (p mod 4) eq 1 };
    return sq2 select 2 else 3;
end intrinsic;

intrinsic LengthOfSumOfSquares( a::RngIntElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;

intrinsic LengthOfSOS( a::FldRatElt ) -> RngIntElt
{
Given a nonzero rational number a, this fuction determines
minimal number of summands needed to express a as a sum of squares. 
It returns 0 if a is not a sum of squares.
}
    return LengthOfSOS( Numerator(a)*Denominator(a) );
end intrinsic;

intrinsic LengthOfSumOfSquares( a::FldRatElt ) -> RngIntElt
{ " } //"
    return LengthOfSOS(a);
end intrinsic;


//////////////////////////////////////////////////
// SolveSumOfSquares

function ssos_len2( a )
// Given an element of length 2, this function returns c1, c2
// such that c1^2 + c2^2 = a

    L := QuadraticField(-1);
    test, v := NormEquation(L, a);
    assert test;
    // convert to coeffs
    v := Eltseq( v[1] );
    c1 := v[1];
    c2 := v[2];
    
    assert c1^2 + c2^2 eq a;
    return [ c1, c2 ];
end function;

function ssos_len3( n )
// Implementation of [ Pollack-Schorn]

    // divide by power of 4
    k := Floor( Valuation(n,2)/2 );
    n div:= 4^k;

    iters := 100;
    bitsize := Ceiling(2 + Log(2,8*n));
    test := false;
    case n mod 8 :
    when 2, 6 :
        while not test do 
            test, p := RandomPrime(bitsize, 2*n-1, 4*n, iters);
            bitsize *:= 2;
            iters *:= 10;
        end while;
        D := (p+1) div n;
    when 1 :
        while not test do 
            test, p := RandomPrime(bitsize, 6*n-1, 8*n, iters);
            bitsize *:= 2;
            iters *:= 10;
        end while;
        D := (p+1) div n;
    when 3 :
        while not test do 
            test, p := RandomPrime(bitsize, (5*n-1) div 2, 4*n, iters);
            bitsize *:= 2;
            iters *:= 10;
        end while;
        D := (2*p+1) div n;
    when 5 :
        while not test do 
            test, p := RandomPrime(bitsize, (3*n-1) div 2, 4*n, iters);
            bitsize *:= 2;
            iters *:= 10;
        end while;
        D := (2*p+1) div n;
    else :
        error "Wrong argument n";
    end case;
    
    a12 := Modsqrt( -D, D*n - 1);
    a22 := D*n - 1;
    a11 := (D + a12^2) div a22;
    
    // matrix
    M := Matrix(ZZ, [[a11, a12, 1], [a12, a22, 0], [1, 0, n]]);
    R, U, r := LLLGram(M);
    
    c := Eltseq( (U^-1)[3] );
    assert &+ [ ci^2 : ci in c ] eq n;
    return [ 2^k*ci : ci in c ];
end function;

/*
function ssos_len4( n )
// Implementation of [ Pollack-Trevino, 2018 ]
    // divide by power of 2
    e := Floor( Valuation(n,2) );
    n_ := n div 2^e;

    // search for a prime p = -1 (mod n) and p = 1 (mod 4)
    m := CRT( [-1, 1], [n_, 4] );
    found := false;
    iters := 100;
    bitsize := Ceiling(2 + Ilog2(8*n_));
    while not found do
        found, p := RandomPrime(bitsize, m, 4*n_, iters);
        bitsize *:= 2;
        iters *:= 10;
    end while;
    
    // decompose p into 2 squares
    c := ssos_len2(p);
    A := c[1];
    B := c[2];
    
    // NO WAY TO COMPUTE GCD IN HURWITZ QUATERNIONS!
    
end function;
*/

function ssos_len4( n )
// Using Simon's function
    V := IsotropicSubspace( DiagonalMatrix(ZZ,[1,1,1,1,-n]) );
    
    c := Eltseq(V.1);
    c := [ c[j]/c[5] : j in [1..4] ];
    assert &+ [ ci^2 : ci in c ] eq n;
    return c;
end function;


intrinsic SolveSumOfSquares( a::RngIntElt ) -> []
{
Let a be a sum of squares in ZZ. Assume that the length of a is n.
This function solves the equation a = x1^2 + ... + xn^2 in ZZ.
}
    require IsSumOfSquares(a) : "The argument must be a sum of squares!";
    
    test, sqrt := IsSquare(a);
    if test then
        return [ sqrt ];
    end if;
    
    l := LengthOfSOS(a);
    if l eq 2 then 
        return ssos_len2(a);
    elif l eq 3 then
        return ssos_len3(a);
    else 
        return ssos_len4(a);
    end if;
end intrinsic;

intrinsic SolveSOS( a::RngIntElt ) -> []
{ " } //"
    return SolveSumOfSquares(a);
end intrinsic;

intrinsic SolveSumOfSquares( a::FldRatElt ) -> []
{
Let a be a sum of squares in QQ. Assume that the length of a is n.
This function solves the equation a = x1^2 + ... + xn^2 in QQ.
}
    require IsSumOfSquares(a) : "The argument must be a sum of squares!";

    d := Denominator(a);
    n := Numerator(a)*d;
    c := SolveSumOfSquares(n);
    c := [ ci/d : ci in c ];
    assert &+ [ ci^2 : ci in c ] eq a;
    return c;
end intrinsic;

intrinsic SolveSOS( a::FldRatElt ) -> []
{ " } //"
    return SolveSumOfSquares(a);
end intrinsic;

